"""
Shared exceptions, logging and pydantic models. Not a lot to report here.

Pydantic data models may contain more classes as validator for fields; for stand-alone models,
use the classes marked with `Model` at the end of the class name.
"""

